#define F_CPU 8000000UL	/* CPU clock in Hertz */

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>


//Vergleichswert fr Dimmer, hhere Zahl=spter einschalten, 180=sofort nach ZC
//Untere Grenze entspricht ca. 300W Leistung bzw. 26 Prozent der Halbwelle
//Tabelle im Flash ablegen, da SRAM sonst voll
const uint8_t Kurve[] PROGMEM = 
{229,227,224,222,220,219,217,215,213,211,209,207,205,203,200,198,195,192,188,180};

volatile uint8_t	triac = 0;		//Ist der Triac eingeschaltet?

volatile uint8_t	timer = 0;		//Timer fr Sekundentakt
volatile uint8_t	timer2 = 0;		//Timer fr Zentelsekundentakt

volatile uint8_t	ocr = 240;		//Startwert mit geringer Leistung, wird nach 1s aktualisiert
volatile uint8_t	adc = 0;		//aktueller Vergleichswert 0...19
volatile uint8_t	adcNEU = 0;		//ausgelesener Vergleichswert 0...19
volatile uint16_t 	adcRAW = 0; 	//Rohwert aus dem ADC

//Ausgnge
#define DIMMER_AUS	PORTB &= ~(1<<PB4);
#define DIMMER_AN	PORTB |= (1<<PB4);

//#define ZC_AUS	PORTB &= ~(1<<PB0);
//#define ZC_AN	PORTB |= (1<<PB0);


/* ADC initialisieren */
void ADC_Init(void)
{
  // die Versorgungsspannung AVcc als Refernz whlen:
  //ADMUX = (1<<REFS0);    
 
  // Bit ADFR ("free running") in ADCSRA steht beim Einschalten
  // schon auf 0, also single conversion
  ADCSRA = (1<<ADPS2)|(1<<ADPS1)|(1<<ADPS0);     // Frequenzvorteiler
  ADCSRA |= (1<<ADEN);                  // ADC aktivieren
 
  /* nach Aktivieren des ADC wird ein "Dummy-Readout" empfohlen, man liest
     also einen Wert und verwirft diesen, um den ADC "warmlaufen zu lassen" */
 
  ADCSRA |= (1<<ADSC);                  // eine ADC-Wandlung 
  while (ADCSRA & (1<<ADSC) ) {         // auf Abschluss der Konvertierung warten
  }
  /* ADCW muss einmal gelesen werden, sonst wird Ergebnis der nchsten
     Wandlung nicht bernommen. */
  (void) ADCW;
}

/* ADC Einzelmessung */
uint16_t ADC_Read( uint8_t channel )
{
  // Kanal waehlen, ohne andere Bits zu beeinfluen
  ADMUX = (ADMUX & ~(0x1F)) | (channel & 0x1F);
  ADCSRA |= (1<<ADSC);            // eine Wandlung "single conversion"
  while (ADCSRA & (1<<ADSC) ) {}  // auf Abschluss der Konvertierung warten
  return ADCW;                    // ADC auslesen und zurckgeben
}

//############################################################################
//ISR vom externen Interrupt (Zero cross detection)
ISR (INT0_vect)
//############################################################################
{
	DIMMER_AUS;
	TCNT0  = 178;								//sync T0 with mains zc
	OCR0B  = ocr; 								//set first compare angle (full brightness)
	TIMSK  |= (1<<OCIE0B)|(1<<TOIE0);			//enable timer compare irq + overflow irq
	TIFR   |= (1<<OCF0B)|(1<<TOV0);				//clear Timer flag
	GIMSK  &= ~(1<<INT0);						//disable self (avoid multi-triggering by noise)
	
}

//############################################################################
//Timer 0 overflow interrupt - erste Halbwelle am Ende, lade zweite Halbwelle
ISR (TIMER0_OVF_vect)
//############################################################################
{
	DIMMER_AUS;
	TCNT0  = 178;				//Nchste Halbwelle wird per Software berechnet
	OCR0B  = ocr;
	TIMSK  |= (1<<OCIE0B);		//enable compare irq
	GIFR   |= (1<<INTF0);		//clear zc flag
	GIMSK  |= (1<<INT0);		//enable int0 - zc irq
	TIMSK  &= ~(1<<TOIE0);		//disable self
}

//############################################################################
//Timer 0 B compare interrupt - Zndzeitpunkt erreicht
ISR (TIMER0_COMPB_vect)
//############################################################################
{
	if (triac<1)
	{
	DIMMER_AN;				//TRIAC feuern
	triac=1;
	OCR0B  = ocr + 2; 		//nach 2 Takten (>0,2ms) diesen Aufruf wiederholen
							// und den Triac ausschalten (nuckelt sonst das
							// Kondensatornetzteil leer)
	} else {
	DIMMER_AUS;
	triac=0;
	TIMSK  &= ~(1<<OCIE0B); //disable self
	}
}


//############################################################################
//Timer 1 overflow interrupt - 0,01s Takt
ISR (TIMER1_OVF_vect)
//############################################################################
{
	timer++;
	timer2++;
	TCNT1  = 178;
}


//------------------------------------------------------------------------------------------------------------
int main(void){

	//########### INIT ##########
	
	/* Timer 0 (8bit) fr Dimmer (eine Halbwelle) 10ms
	Gewnscht: 0.01 s
	Am nchsten gelegen: 0.009984 s
	Verwendeter prescaler: 1024
	Verwendeter preload: 178
	-> 255-178 = 77 Stufen zum Dimmen
	*/
	TCCR0B |= (1<<CS02) | (1<<CS00);
	TCNT0 = 178;
	TIMSK |= (1<<TOIE0);	//Timer Overflow Interrupt 0 enable
	
	/* Timer 1 (8bit) fr Zeit 10ms
	Gewnscht: 0.01 s
	Am nchsten gelegen: 0.009984 s
	Verwendeter prescaler: 1024
	Verwendeter preload: 178
	*/
	TCCR1 |= (1<<CS10) | (1<<CS11) | (1<<CS13);
	//TCCR1 |= (1<<CS10) | (1<<CS11) | (1<<CS12);
	TCNT1 = 178;//131;
	TIMSK |= (1<<TOIE1);	//Timer Overflow Interrupt 0 enable


	//INT0 Nulldurchgang
	MCUCR = (1<<ISC01); 	//fallende Flanke
	GIMSK = (1<<INT0); 		//Ext. Interrup 0 enable
	GIFR  = (1<<INTF0);		//clear INT1 flag

	//ADC init
	ADC_Init();

	//Globale Interrupts Enable	
	sei();

	while(1){

		if (timer2==10) //Zentelsekunde
			{
			//Schrittweise erhhung soll das System stabiler machen
			if (adcNEU>adc) adc++;
			if (adcNEU<adc) adc--;
			timer2=0;
			}

		if (timer == 100)	//eine Sekunde
			{
			adcRAW = ADC_Read(3); //Poti an Pin PB3 auslesen
			adcNEU = adcRAW / 53; //auf 20 Stufen (1024/53=19,xx) herunterrechnen
			if (adcNEU>19) adcNEU=19; //Maximum deckeln
			ocr = pgm_read_word(Kurve + adc);
			timer = 0;
			}

	}//ENDE While(1)

}
// EOF----------------------------------------------------------------------------------------------------------
